from datetime import datetime
from django.db import models
from django.conf import settings

from response.core.models import Incident, ExternalUser, GetOrCreateSlackExternalUser

class UserStats(models.Model):
    user = models.ForeignKey(ExternalUser, on_delete=models.CASCADE, blank=False, null=False)
    incident = models.ForeignKey(Incident, on_delete=models.CASCADE, blank=False, null=False)

    join_time = models.DateTimeField(null=True)
    message_count = models.IntegerField(default=0)

    class Meta:
        unique_together = ("incident", "user")

    @staticmethod
    def increment_message_count(incident, user_id):
        name = settings.SLACK_CLIENT.get_user_profile(user_id)['name']
        user = GetOrCreateSlackExternalUser(external_id=user_id, display_name=name)

        user_stats, created = UserStats.objects.get_or_create(incident=incident, user=user)

        if created:
            user_stats.join_time = datetime.now()

        user_stats.message_count += 1
        user_stats.save()

    def __str__(self):
        return f"{self.user.display_name} - {self.incident}"



from django.apps import AppConfig


class ResponseConfig(AppConfig):
    name = 'response'

    def ready(self):
        from .slack import (settings,
                            signals,
                            action_handlers,
                            event_handlers,
                            incident_commands,
                            incident_notifications,
                            dialog_handlers)



import after_response

import logging
logger = logging.getLogger(__name__)

DIALOG_HANDLERS = {}


def dialog_handler(callback_id, func=None):
    def _wrapper(fn):
        DIALOG_HANDLERS[callback_id] = fn
    if func:
        return _wrapper(func)
    return _wrapper

def remove_dialog_handler(callback_id):
    DIALOG_HANDLERS.pop(callback_id, None)

@after_response.enable
def handle_dialog(payload):
    callback_id = payload['callback_id']
    if callback_id not in DIALOG_HANDLERS:
        logger.error(f"Can't find handler for dialog id {callback_id}")
        return

    callback = DIALOG_HANDLERS[callback_id]

    user_id = payload['user']['id']
    channel_id = payload['channel']['id']
    submission = payload['submission']
    response_url = payload['response_url']
    state = payload['state']

    callback(user_id, channel_id, submission, response_url, state)



from .action_handler import *
from .event_handler import *
from .headline_post_action import *
from .incident_command import *
from .keyword_handler import *
from .incident_notification import *
from .dialog_handler import *



from response.core.models.incident import Incident
from response.slack.models.comms_channel import CommsChannel

import logging
logger = logging.getLogger(__name__)

KEYWORD_HANDLERS = {}


def keyword_handler(keywords, func=None):
    def _wrapper(fn):
        for keyword in keywords:
            KEYWORD_HANDLERS[keyword] = fn
        return fn
    if func:
        return _wrapper(func)
    return _wrapper


def handle_keywords(payload):
    text = payload.get('text', '')
    user = payload.get('user', '')
    ts = payload.get('ts', '')
    channel = payload.get('channel', '')

    comms_channel = CommsChannel(channel_id=channel)

    for keyword, handler in KEYWORD_HANDLERS.items():
        if keyword.lower() in text.lower():
            handler(comms_channel, user, text, ts)



from rest_framework import serializers
from rest_framework.decorators import action

from response.core.models.incident import Incident
from response.core.models.action import Action
from response.core.models.user_external import ExternalUser

from django.contrib.auth.models import User


class ExternalUserSerializer(serializers.ModelSerializer):
    class Meta:
        model = ExternalUser
        fields = ("app_id", "external_id", "display_name")


class ActionSerializer(serializers.ModelSerializer):
    # Serializers define the API representation.
    incident = serializers.PrimaryKeyRelatedField(
        queryset=Incident.objects.all(), required=False
    )
    user = serializers.PrimaryKeyRelatedField(
        queryset=ExternalUser.objects.all(), required=False
    )

    class Meta:
        model = Action
        fields = ("pk", "details", "done", "incident", "user")


class IncidentSerializer(serializers.ModelSerializer):
    reporter = ExternalUserSerializer(read_only=True)
    lead = ExternalUserSerializer()

    class Meta:
        model = Incident
        fields = (
            "action_set",
            "end_time",
            "impact",
            "lead",
            "pk",
            "report",
            "report_time",
            "reporter",
            "severity",
            "start_time",
            "summary",
        )

    def update(self, instance, validated_data):
        instance.end_time = validated_data.get("end_time", instance.end_time)
        instance.impact = validated_data.get("impact", instance.impact)

        new_lead = validated_data.get("lead", None)
        if new_lead:
            instance.lead = ExternalUser.objects.get(
                display_name=new_lead["display_name"],
                external_id=new_lead["external_id"],
            )

        instance.report = validated_data.get("report", instance.report)
        instance.start_time = validated_data.get("start_time", instance.start_time)
        instance.summary = validated_data.get("summary", instance.summary)
        instance.severity = validated_data.get("severity", instance.severity)

        instance.save()
        return instance



from rest_framework import pagination, viewsets

from response.core.models.incident import Incident
from response.core.models.action import Action
from response.core.models.user_external import ExternalUser
from response.core import serializers

from datetime import datetime
from calendar import monthrange


class ExternalUserViewSet(viewsets.ModelViewSet):
    # ViewSets define the view behavior.
    queryset = ExternalUser.objects.all()
    serializer_class = serializers.ExternalUserSerializer


class ActionViewSet(viewsets.ModelViewSet):
    # ViewSets define the view behavior.
    queryset = Action.objects.all()
    serializer_class = serializers.ActionSerializer


class IncidentViewSet(viewsets.ModelViewSet):
    """
    Allows getting a list of Incidents (sorted by report time from newest to
    oldest), and updating existing ones.

    Note that Incidents can only be created via the Slack workflow.
    """

    queryset = Incident.objects.order_by("-report_time")

    serializer_class = serializers.IncidentSerializer
    pagination_class = pagination.LimitOffsetPagination



from django.conf.urls import url, include
from rest_framework import routers

from response.core.views import IncidentViewSet, ActionViewSet, ExternalUserViewSet

# Routers provide an easy way of automatically determining the URL conf.
router = routers.DefaultRouter()
router.register(r'incidents', IncidentViewSet, basename='incident')
router.register(r'actions', ActionViewSet)
router.register(r'ExternalUser', ExternalUserViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browsable API.
urlpatterns = [
    url(r'^', include(router.urls)),
]



import json
from datetime import datetime

from django.conf import settings

from response.slack.settings import INCIDENT_EDIT_DIALOG, INCIDENT_REPORT_DIALOG
from response.core.models.incident import Incident
from response.slack.models import HeadlinePost, CommsChannel, ExternalUser, GetOrCreateSlackExternalUser
from response.slack.decorators import dialog_handler
from response.slack.client import channel_reference

import logging
logger = logging.getLogger(__name__)


@dialog_handler(INCIDENT_REPORT_DIALOG)
def report_incident(user_id: str, channel_id: str, submission: json, response_url: str, state: json):
    report = submission['report']
    summary = submission['summary']
    impact = submission['impact']
    lead_id = submission['lead']
    severity = submission['severity']

    name = settings.SLACK_CLIENT.get_user_profile(user_id)['name']
    reporter = GetOrCreateSlackExternalUser(external_id=user_id, display_name=name)

    lead = None
    if lead_id:
        lead_name = settings.SLACK_CLIENT.get_user_profile(lead_id)['name']
        lead = GetOrCreateSlackExternalUser(external_id=lead_id, display_name=lead_name)

    Incident.objects.create_incident(
        report=report,
        reporter=reporter,
        report_time=datetime.now(),
        summary=summary,
        impact=impact,
        lead=lead,
        severity=severity,
    )

    incidents_channel_ref = channel_reference(settings.INCIDENT_CHANNEL_ID)
    text = f"Thanks for raising the incident 🙏\n\nHead over to {incidents_channel_ref} to complete the report and/or help deal with the issue"
    settings.SLACK_CLIENT.send_ephemeral_message(channel_id, user_id, text)


@dialog_handler(INCIDENT_EDIT_DIALOG)
def edit_incident(user_id: str, channel_id: str, submission: json, response_url: str, state: json):
    report = submission['report']
    summary = submission['summary']
    impact = submission['impact']
    lead_id = submission['lead']
    severity = submission['severity']

    lead = None
    if lead_id:
        lead_name = settings.SLACK_CLIENT.get_user_profile(lead_id)['name']
        lead = GetOrCreateSlackExternalUser(external_id=lead_id, display_name=lead_name)

    try:
        incident = Incident.objects.get(pk=state)

        # deliberately update in this way the post_save signal gets sent
        # (required for the headline post to auto update)
        incident.report = report
        incident.summary = summary
        incident.impact = impact
        incident.lead = lead
        incident.severity = severity
        incident.save()

    except Incident.DoesNotExist:
        logger.error(f"No incident found for pk {state}")



from datetime import datetime
from django.db import models
from django.conf import settings

from response.core.models import Incident, ExternalUser, GetOrCreateSlackExternalUser


class PinnedMessageManager(models.Manager):
    def add_pin(self, incident, message_ts, author_id, text):
        name = settings.SLACK_CLIENT.get_user_profile(author_id)['name']
        author = GetOrCreateSlackExternalUser(external_id=author_id, display_name=name)

        PinnedMessage.objects.get_or_create(
            incident=incident,
            message_ts=message_ts,
            defaults={
                'author': author,
                'text': text,
                'timestamp': datetime.fromtimestamp(float(message_ts)),
            }
        )

    def remove_pin(self, incident, message_ts):
        PinnedMessage.objects.filter(
            incident=incident,
            message_ts=message_ts,
        ).delete()


class PinnedMessage(models.Model):
    incident = models.ForeignKey(Incident, on_delete=models.CASCADE)
    author = models.ForeignKey(ExternalUser, on_delete=models.PROTECT, blank=False, null=False)
    message_ts = models.CharField(max_length=50, blank=False, null=False)
    text = models.TextField()
    timestamp = models.DateTimeField()

    objects = PinnedMessageManager()

    def __str__(self):
        return f"{self.text}"



from django.contrib.auth.models import User
import factory
from faker import Factory


from response.core.models import ExternalUser

faker = Factory.create()


class UserFactory(factory.DjangoModelFactory):
    class Meta:
        model = User

    username = factory.LazyFunction(faker.user_name)
    password = factory.LazyFunction(faker.password)


class ExternalUserFactory(factory.DjangoModelFactory):
    class Meta:
        model = ExternalUser

    app_id = "slack"
    display_name = factory.LazyFunction(faker.name)
    external_id = factory.LazyFunction(lambda: "U" + faker.word())
    owner = factory.SubFactory("tests.factories.UserFactory")



from response.core.models import Incident
from response.slack.models import CommsChannel
from response.slack .decorators import recurring_notification, single_notification


@recurring_notification(interval_mins=5, max_notifications=5)
def remind_severity(incident: Incident):
    try:
        comms_channel = CommsChannel.objects.get(incident=incident)
        if not incident.severity:
            comms_channel.post_in_channel("🌤️ This incident doesn't have a severity.  Please set one with `@incident severity ...`")
    except CommsChannel.DoesNotExist:
        pass


@recurring_notification(interval_mins=2, max_notifications=5)
def remind_incident_lead(incident: Incident):
    try:
        comms_channel = CommsChannel.objects.get(incident=incident)
        if not incident.lead:
            comms_channel.post_in_channel("👩‍🚒 This incident hasn't got a lead.  Please set one with `@incident lead ...`")
    except CommsChannel.DoesNotExist:
        pass

@recurring_notification(interval_mins=720, max_notifications=6)
def remind_close_incident(incident: Incident):
    try:
        comms_channel = CommsChannel.objects.get(incident=incident)
        if not incident.is_closed():
            comms_channel.post_in_channel(":timer_clock: This incident has been running a long time.  Can it be closed now?  Remember to pin important messages in order to create the timeline.")
    except CommsChannel.DoesNotExist:
        pass


from django.db.models.signals import post_save
from django.core.signals import request_finished
from django.dispatch import receiver

from response.core.models import Incident
from response.slack.models import HeadlinePost

from time import sleep


@receiver(post_save, sender=Incident)
def update_headline_after_incident_save(sender, instance, **kwargs):
    """
    Reflect changes to incidents in the headline post

    Important: this is called in the synchronous /incident flow so must remain fast (<2 secs)

    """
    try:
        headline_post = HeadlinePost.objects.get(
            incident=instance
        )
        headline_post.update_in_slack()

    except HeadlinePost.DoesNotExist:
        headline_post = HeadlinePost.objects.create_headline_post(
            incident=instance
        )


@receiver(post_save, sender=HeadlinePost)
def update_headline_after_save(sender, instance, **kwargs):
    """
    Reflect changes to headline posts in slack

    """
    instance.update_in_slack()
